/* TIGRE implementation:  HEAP.C */
/* Heap management and garbage collection */
/* Uses stop-and-copy garbage collector */

/* (C) Copyright 1989   Philip Koopman Jr.  */
/*  Last update:  2/13/89   */
/*  Test code version -- not for distribution  */

#include "reduce.h"
#include "heap.h"

   Celltype *root_save ;              /* root of evalution tree */
   Celltype *next_free_node ;         /* heap free list pointer */
   Celltype *heap_start, *heap_end ;  /* boundaries of heap memory */
   Celltype *from_heap, *to_heap ;  /* two heap buffers */

   Celltype **spine_start,**spine_end ;  /* boundaries of spine memory */
   Celltype **spine_stack ;    /* points to topmost element on spine stack */
   Celltype *spine_mem[SPINESIZE+100];    /* spine stack storage */
   Celltype **fixup_start ;  /* starting point for stack fixup */

   int using_first_heap ; /* true flag with first_heap is current */
   int gc ;      /* counts number of garbage collections used */


void zapflags()  /* set all flags in heap arrays to false */
  { register int i;
    register Celltype *zapptr ;
#if VAX_ASM
    zapptr = from_heap ;
    for (i = 0 ; i <= (HEAPSIZE) ; i++ )
       { zapptr->value.comb = UNMARKED ;
         zapptr += 1 ;
         if(IS_FORWARDED(zapptr->value.child))
                 zapptr->value.child = UNFORWARD(zapptr->value.child) ;
         zapptr += HEAP_STRIDE - 1 ;
       }
    zapptr = to_heap ;
    for (i = 0 ; i <= (HEAPSIZE) ; i++ )
       { zapptr->value.comb = UNMARKED ;
         zapptr += 1 ;
         if(IS_FORWARDED(zapptr->value.child))
                 zapptr->value.child = UNFORWARD(zapptr->value.child) ;
         zapptr += HEAP_STRIDE - 1 ;
       }
#else
    zapptr = from_heap+1 ;
    for (i = 0 ; i <= (HEAPSIZE) ; i++ )
       { if(IS_FORWARDED(zapptr->value.child))
                 zapptr->value.child = UNFORWARD(zapptr->value.child) ;
         zapptr += HEAP_STRIDE ;
       }
    zapptr = to_heap+1 ;
    for (i = 0 ; i <= (HEAPSIZE) ; i++ )
       { if(IS_FORWARDED(zapptr->value.child))
                 zapptr->value.child = UNFORWARD(zapptr->value.child) ;
         zapptr += HEAP_STRIDE ;
       }
#endif
  }

void init_heap()  /* init heap memory -- leaves first xx nodes free */
{  int i ;  /* loop counter */
#if DEBUG_HEAP
    printf(" init_heap");
#endif
  heap_start = to_heap ;
  heap_end = heap_start+(HEAPSIZE*HEAP_STRIDE) ;
  /* set next_free_node halfway through to allow room for program */
  next_free_node = heap_start+(HEAP_STRIDE*PROGSIZE)+1 ;

  spine_start = &spine_mem[2] ;
  spine_end = &spine_mem[SPINESIZE-2] ;
  spine_stack = spine_end ;
  fixup_start = spine_end - 1 ; 
  zapflags();

  for ( i = 0 ; i <= MAX_TOKEN ; i++ )
    { i_counts[i] = 0 ; }
  i_heap = 0 ;
  ready_for_longjmp = FALSE ;
}

Celltype *newnode_continue(top_spine)
  Celltype **top_spine;
{ /* ran out of heap -- do garbage collection */
  register Celltype *mynode ;
  collect_garbage(top_spine);
  mynode = next_free_node ;
  next_free_node = mynode + HEAP_STRIDE ;
  return(mynode) ;
}

Celltype *newnode(top_spine)
 Celltype **top_spine;
{ /* allocate a node unless free node pointer goes too far */
  register Celltype *mynode ;
  mynode = next_free_node ;
  next_free_node = mynode + HEAP_STRIDE ;
  if( mynode < heap_end)  return(mynode) ;
  return(newnode_continue(top_spine));
}

void newnode2_continue(top_spine, nodea, nodeb)
 Celltype **top_spine, **nodea, **nodeb;
{ register Celltype *temp ;
  /* ran out of heap -- do garbage collection */
  collect_garbage(top_spine);
  temp = next_free_node ;
  *nodea = temp ;
  temp = temp + HEAP_STRIDE ;
  *nodeb = temp ;
  next_free_node = temp + HEAP_STRIDE ;
  return ;
}
void newnode2(top_spine, nodea, nodeb)
  Celltype **top_spine, **nodea, **nodeb;
{ register Celltype *temp ;
  temp = next_free_node ;
  *nodea = temp ;
  temp = temp + HEAP_STRIDE ;
  *nodeb = temp ;
  next_free_node = temp + HEAP_STRIDE ;
  if( temp < heap_end)  return ;
  newnode2_continue(top_spine,nodea,nodeb);
  return;
}

void newnode3_continue(top_spine, nodea, nodeb, nodec)
 Celltype **top_spine, **nodea, **nodeb, **nodec;
{ register Celltype *temp ;
  /* ran out of heap -- do garbage collection */
  collect_garbage(top_spine);
  temp = next_free_node ;
  *nodea = temp ;
  temp = temp + HEAP_STRIDE ;
  *nodeb = temp ;
  temp = temp + HEAP_STRIDE ;
  *nodec = temp ;
  next_free_node = temp + HEAP_STRIDE ;
  return ;
}
void newnode3(top_spine, nodea, nodeb, nodec)
  Celltype **top_spine, **nodea, **nodeb, **nodec;
{ register Celltype *temp ;
  temp = next_free_node ;
  *nodea = temp ;
  temp = temp + HEAP_STRIDE ;
  *nodeb = temp ;
  temp = temp + HEAP_STRIDE ;
  *nodec = temp ;
  next_free_node = temp + HEAP_STRIDE ;
  if( temp < heap_end)  return ;
  newnode3_continue(top_spine,nodea, nodeb, nodec) ;
  return;
}


Celltype *newnoden_continue(n, top_spine)
  int n;
  Celltype **top_spine;
{ /* ran out of heap -- do garbage collection */
  register Celltype *mynode ;
  collect_garbage(top_spine);
  mynode = next_free_node ;
  next_free_node = mynode + HEAP_STRIDE * n ;
  return(mynode) ;
}


void collect_garbage(top_spine)
  Celltype **top_spine;
{ UNIX_REGISTER Celltype from_ptr ; /* scratch pointers for the copying */
  UNIX_REGISTER Celltype temp ;
  register Celltype *scanned ;
  UNIX_REGISTER Celltype *unscanned ;
  UNIX_REGISTER Celltype lit_value ;  /* holding buffer for LIT compare */
#if DEBUG_HEAP
     printf(" collect_garbage");
#endif
  gc++ ;
  lit_value = do_lit_value ;
#if DEBUG_DUMP
        printf("\n\n *** graph dump %X %X ***",heap_start,heap_end);
        zapflags();
        dumpgraph(root_save,0);
        zapflags();
        printf("\n *** end of graph dump ***\n");
#endif
#if DEBUG_HEAP
   printf("gc#%d ",gc);
#endif
  /* exchange the role of the heaps, since to_heap was previously in use */
  temp.value.ptr = to_heap ;
  to_heap = from_heap ;
  from_heap = temp.value.ptr ;
  heap_start = to_heap ;
  heap_end = heap_start+(HEAPSIZE*HEAP_STRIDE) ;
  scanned = heap_start + 1 + HEAP_STRIDE ;
  unscanned = scanned + HEAP_STRIDE ;

  /* establish the root node as the base case for the to heap */
  scanned->value = root_save->value ;
  (scanned+1)->value = (root_save+1)->value ;
  root_save = scanned ;

  /* scan through the to space until all referenced from nodes copied */
  while ( scanned < unscanned )
  {
#if DEBUG_HEAP
    printf("\n node @ %X",scanned);
#endif
       /* ------ copy this section for second half */
    from_ptr.value.ptr = scanned->value.child ;  /* target of LHS pointer */
#if DEBUG_HEAP
    printf("\n --- LHS = %X",from_ptr.value.ptr);
#endif
    if (IS_PTR(from_ptr.value.ptr))
    { /* It's a pointer. */
      from_ptr.value.ptr = UNFUDGE(from_ptr.value.ptr) ;
      temp = *from_ptr.value.ptr ;
      if ( IS_FORWARDED(temp.value.ptr) )
      { /* the target cell is already resident in the heap */
#if DEBUG_HEAP
        printf("\n target in heap at %X",UNFORWARD(temp.value.ptr));
#endif
        scanned->value.child = UNFORWARD(temp.value.ptr) ;
      }
      else
      { /* the target cell must be moved */
#if DEBUG_HEAP
        printf("\n target moved to    %X",unscanned);
#endif
        scanned->value.child = FUDGE(unscanned) ;
        *(unscanned) = temp ;
        *(unscanned+1) = *(from_ptr.value.ptr+1) ;
        from_ptr.value.ptr->value.ptr =  MAKE_FORWARD(FUDGE(unscanned)) ;
        unscanned += HEAP_STRIDE ;  /* skip the mark word */
      }
    }
       /* ------ end of copy section */
    else
    { /* It's a combinator */
      if ( scanned->value.comb == lit_value.value.comb )
      { /* if the LHS was a LIT combinator, then skip RHS */
#if DEBUG_HEAP
        printf("\n LIT combinator\n");
#endif
        scanned += HEAP_STRIDE;
        continue ;
      }
    }
    /* now do the RHS of the scanned node */
    scanned += 1;
    /* ---- copied from above condition clause */
    from_ptr.value.ptr = scanned->value.child ;  /* target of LHS pointer */
#if DEBUG_HEAP
    printf("\n --- RHS = %X",from_ptr.value.ptr);
#endif
    if (IS_PTR(from_ptr.value.ptr))
    { /* It's a pointer. */
      from_ptr.value.ptr = UNFUDGE(from_ptr.value.ptr) ;
      temp = *from_ptr.value.ptr ;
      if ( IS_FORWARDED(temp.value.ptr) )
      { /* the target cell is already resident in the heap */
#if DEBUG_HEAP
        printf("\n target in heap at %X",UNFORWARD(temp.value.ptr));
#endif
        scanned->value.child = UNFORWARD(temp.value.ptr) ;
      }
      else
      { /* the target cell must be moved */
#if DEBUG_HEAP
        printf("\n target moved to    %X",unscanned);
#endif
        scanned->value.child = FUDGE(unscanned) ;
        *(unscanned) = temp ;
        *(unscanned+1) = *(from_ptr.value.ptr+1) ;
        from_ptr.value.ptr->value.ptr =  MAKE_FORWARD(FUDGE(unscanned)) ;
        unscanned += HEAP_STRIDE ;  /* skip the mark word */
      }
    }

    /* ------ end of copy */
    scanned += (HEAP_STRIDE - 1);  /* skip the mark for the next cell */
  } ;

  next_free_node = unscanned ;
  /* do the return stack */
#if DEBUG_HEAP
  printf("rs ");
#endif

  if ( (!ready_for_longjmp)
  && (top_spine != spine_end ))  fixup_stack(top_spine); 
#if DEBUG_DUMP
        printf("\n\n *** graph dump %X %X ***",heap_start,heap_end);
        zapflags();
        dumpgraph(root_save,0);
        zapflags();
        printf("\n *** end of graph dump ***\n");
#endif
#if DEBUG_HEAP
  printf("done... \n"); 
  printf("\nGC  %X used, sp = %X  ",(next_free_node - heap_start),
               (fixup_start -  top_spine) );
  if ( ready_for_longjmp)  printf (" -- longjmp") ;
#endif

  if( next_free_node < ( heap_end - 256 ) ) 
   { 
     if(ready_for_longjmp)
#if UNIX
          { _longjmp(env,1);}
#else
          {  longjmp(env,1);}
#endif
     else {  return ;}
   }
  printf("\n\n ***** heap space overflow (burp!) **** \n\n");
  exit(-1);
}

void fixup_stack(top_spine)
  Celltype **top_spine;
{ UNIX_REGISTER Celltype temp ;
  UNIX_REGISTER Celltype **temp_spine ;

#if DEBUG_HEAP
  printf("\nstack fixup  start:%X  end:%X\n",fixup_start,top_spine);
#endif

/* fix up the stack for garbage collection */
  temp_spine = top_spine ; /* skip over saved pointer to root_save */
  while( temp_spine <= fixup_start )
  { /* copy in forwarded address */
#if DEBUG_HEAP
    printf("\nmoving %X (%X)",temp_spine,*temp_spine);
#endif
    temp.value.ptr = *temp_spine ;
    if ( temp.value.ptr < from_heap ) /* check for subr call */  
       { /* skip a data value */
#if DEBUG_HEAP
         printf("...skip... %X %X",*temp_spine,*(temp_spine+1));
#endif
         temp_spine += 2;
         continue ;
       }
    temp = *((*temp_spine)-1) ;  /* get forwarding address */
#if 1+DEBUG_HEAP
    if ( ! IS_FORWARDED(temp.value.ptr) )
       printf("\n bad stack forwarding value=%X  address=%X",
                               temp.value.ptr,*temp_spine);
#endif
    *temp_spine = UNFUDGE(UNFORWARD(temp.value.ptr)) + 1 ;
    temp_spine++ ;
  }
#if DEBUG_HEAP
  printf("\ndone\n\n");
#endif
}
